# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the ssp_generator module."""

import argparse
import pathlib

from _pytest.monkeypatch import MonkeyPatch

from tests import test_utils
from tests.test_utils import FileChecker, setup_for_ssp, setup_for_ssp_fedramp

import trestle.core.generators as gens
import trestle.core.generic_oscal as generic
import trestle.oscal.profile as prof
import trestle.oscal.ssp as ossp
from trestle.common import const, file_utils, list_utils
from trestle.common.model_utils import ModelUtils
from trestle.core.commands.author.ssp import SSPAssemble, SSPFilter, SSPGenerate
from trestle.core.control_context import ContextPurpose, ControlContext
from trestle.core.control_reader import ControlReader
from trestle.core.markdown.markdown_api import MarkdownAPI
from trestle.core.models.file_content_type import FileContentType
from trestle.core.profile_resolver import ProfileResolver

prof_name = 'comp_prof'
ssp_name = 'my_ssp'
cat_name = 'nist_cat'


def confirm_control_contains(trestle_dir: pathlib.Path, control_id: str, part_label: str, seek_str: str) -> bool:
    """Confirm the text is present in the control markdown in the correct part."""
    control_dir = trestle_dir / ssp_name / control_id.split('-')[0]
    md_file = control_dir / f'{control_id}.md'
    context = ControlContext.generate(ContextPurpose.SSP, False, trestle_dir, trestle_dir)
    _, comp_dict = ControlReader.read_control_info_from_md(md_file, context)
    for label_dict in comp_dict.values():
        if part_label in label_dict:
            prose = label_dict[part_label].prose
            if seek_str in prose:
                return True
    return False


part_a_text = """## Implementation for part a.

### comp_aa

statement prose for part a. from comp aa

#### Rules:

  - comp_rule_aa_1

#### Implementation Status: partial

### comp_ab

<!-- Add control implementation description here for item a. -->

#### Rules:

  - comp_rule_ab_1

#### Implementation Status: partial

______________________________________________________________________"""


def test_ssp_generate(tmp_trestle_dir: pathlib.Path) -> None:
    """Test the ssp generator."""
    args, _ = setup_for_ssp(tmp_trestle_dir, prof_name, ssp_name)

    ssp_cmd = SSPGenerate()
    assert ssp_cmd._run(args) == 0
    md_dir = tmp_trestle_dir / ssp_name
    ac_dir = md_dir / 'ac'
    ac_1 = ac_dir / 'ac-1.md'
    assert ac_1.exists()

    md_api = MarkdownAPI()
    header, tree = md_api.processor.process_markdown(ac_1)
    assert header[const.TRESTLE_GLOBAL_TAG][const.SORT_ID] == 'ac-01'
    assert header[const.COMP_DEF_RULES_PARAM_VALS_TAG]['comp_aa'][0] == {
        'name': 'shared_param_1', 'values': ['shared_param_1_aa_opt_1']
    }

    node = tree.get_node_for_key('## Implementation for part a.')
    assert node.content.raw_text == part_a_text

    fc = FileChecker(md_dir)

    assert ssp_cmd._run(args) == 0

    assert fc.files_unchanged()

    assert ssp_cmd._run(args) == 0

    assert fc.files_unchanged()


def test_ssp_generate_no_cds(tmp_trestle_dir: pathlib.Path) -> None:
    """Test the ssp generator with no comp defs."""
    args, _ = setup_for_ssp(tmp_trestle_dir, prof_name, ssp_name)

    args.compdefs = None
    ssp_cmd = SSPGenerate()
    assert ssp_cmd._run(args) == 0
    md_dir = tmp_trestle_dir / ssp_name
    ac_1 = md_dir / 'ac/ac-1.md'
    assert ac_1.exists()
    at_2 = md_dir / 'at/at-2.md'
    assert at_2.exists()

    md_api = MarkdownAPI()
    header, tree = md_api.processor.process_markdown(ac_1)
    assert header[const.TRESTLE_GLOBAL_TAG][const.SORT_ID] == 'ac-01'

    node = tree.get_node_for_key(const.SSP_MD_IMPLEMENTATION_QUESTION, False)
    assert len(node.subnodes) == 1
    assert len(node.subnodes[0].subnodes) == 1
    assert node.subnodes[0].key == '### This System'
    assert node.subnodes[0].subnodes[0].key == '#### Implementation Status: planned'

    fc = FileChecker(md_dir)

    assert ssp_cmd._run(args) == 0

    assert fc.files_unchanged()


def test_ssp_failures(tmp_trestle_dir: pathlib.Path) -> None:
    """Test ssp failure modes."""
    ssp_cmd = SSPGenerate()

    # bad yaml
    yaml_path = test_utils.YAML_TEST_DATA_PATH / 'bad_simple.yaml'
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir,
        profile=prof_name,
        output=ssp_name,
        verbose=0,
        yaml_header=str(yaml_path),
        overwrite_header_values=False
    )
    assert ssp_cmd._run(args) == 1

    # test missing profile
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir,
        profile='foo',
        output=ssp_name,
        verbose=0,
        overwrite_header_values=False,
        yaml_header=None
    )
    assert ssp_cmd._run(args) == 1


def test_ssp_generate_with_yaml_header(tmp_trestle_dir: pathlib.Path) -> None:
    """Test the ssp generator with yaml header."""
    args, _ = setup_for_ssp(tmp_trestle_dir, prof_name, ssp_name, True)
    ssp_cmd = SSPGenerate()
    assert ssp_cmd._run(args) == 0
    ac_dir = tmp_trestle_dir / (ssp_name + '/ac')
    ac_1 = ac_dir / 'ac-1.md'
    assert ac_1.exists()
    assert ac_1.stat().st_size > 1000

    # confirm content from the cli yaml header is now in the header
    md_api = MarkdownAPI()
    header, tree = md_api.processor.process_markdown(ac_1)
    assert tree is not None
    assert header[const.TRESTLE_GLOBAL_TAG][const.SORT_ID] == 'ac-01'
    assert header['control-origination'][0] == 'Service Provider Corporate'


def test_ssp_generate_header_edit(tmp_trestle_dir: pathlib.Path) -> None:
    """Test ssp generate does not overwrite header edits."""
    # always start by creating the markdown with the yaml header
    args, _ = setup_for_ssp(tmp_trestle_dir, prof_name, ssp_name, True)
    ssp_cmd = SSPGenerate()

    cli_yaml_header = args.yaml_header
    args.yaml_header = None

    # first generate with no yaml header
    assert ssp_cmd._run(args) == 0

    ac_dir = tmp_trestle_dir / (ssp_name + '/ac')
    ac_1 = ac_dir / 'ac-1.md'

    md_api = MarkdownAPI()
    header, tree = md_api.processor.process_markdown(ac_1)
    assert tree is not None
    # confirm info from the yaml header is not present
    assert 'control-origination' not in header
    assert 'label' not in header['x-trestle-set-params']['ac-1_prm_5']

    # generate again with header but do not overwrite header values
    args.yaml_header = cli_yaml_header
    assert ssp_cmd._run(args) == 0

    # confirm new items were added from yaml but not when the same key was alread present (values not updated)
    header, tree = md_api.processor.process_markdown(ac_1)
    assert 'control-origination' in header
    assert header['x-trestle-set-params']['ac-1_prm_5']['label'] == 'meetings cancelled from cli yaml'

    # generate again with header and DO overwrite header values
    args.overwrite_header_values = True
    assert ssp_cmd._run(args) == 0

    # confirm values was now changed
    header, tree = md_api.processor.process_markdown(ac_1)
    assert 'control-origination' in header
    assert header['x-trestle-set-params']['ac-1_prm_5']['values'] == 'new values from cli yaml'
    assert header['x-trestle-set-params']['ac-1_prm_5']['label'] == 'meetings cancelled from cli yaml'

    # edit the header by adding a list item and removing a value
    assert file_utils.insert_text_in_file(ac_1, 'System Specific', '  - My new edits\n')
    assert test_utils.delete_line_in_file(ac_1, 'Corporate')

    # tell it not to add the yaml header
    args.yaml_header = None

    assert ssp_cmd._run(args) == 0
    header, tree = md_api.processor.process_markdown(ac_1)

    co = header['control-origination']
    assert co[0] == 'Service Provider System Specific'
    assert co[1] == 'My new edits'
    assert len(co) == 2

    # tell it to add the yaml header but not overwrite header values
    args.yaml_header = cli_yaml_header
    args.overwrite_header_values

    assert ssp_cmd._run(args) == 0
    header, tree = md_api.processor.process_markdown(ac_1)

    # confirm the extra list item from the cli yaml header is added to the list
    co = header['control-origination']
    assert co[2] == 'Service Provider Corporate'
    assert len(co) == 3


def test_ssp_generate_with_inheritance(tmp_trestle_dir: pathlib.Path) -> None:
    """Test ssp-generate with inheritance view."""
    args, _ = setup_for_ssp(tmp_trestle_dir, prof_name, ssp_name, False, 'leveraged_ssp')
    ssp_cmd = SSPGenerate()
    assert ssp_cmd._run(args) == 0

    # Test output for each type of file

    # Find export files under This System
    this_system_dir = tmp_trestle_dir / ssp_name / const.INHERITANCE_VIEW_DIR / 'This System'

    expected_uuid = '11111111-0000-4000-9009-001001002001'
    ac_21 = this_system_dir / 'ac-2.1'
    test_provided = ac_21 / f'{expected_uuid}.md'
    assert test_provided.exists()

    # confirm content in yaml header
    md_api = MarkdownAPI()
    header, tree = md_api.processor.process_markdown(test_provided)
    assert tree is not None

    comp_header_value = header[const.TRESTLE_LEVERAGING_COMP_TAG]
    assert comp_header_value == [{'name': 'REPLACE_ME'}]
    assert header[const.TRESTLE_STATEMENT_TAG][const.PROVIDED_UUID] == expected_uuid

    expected_provided = """# Provided Statement Description

Consumer-appropriate description of what may be inherited.

In the context of the application component in satisfaction of AC-2.1."""

    # Confirm markdown content
    node = tree.get_node_for_key(const.PROVIDED_STATEMENT_DESCRIPTION, False)
    assert node.content.raw_text == expected_provided

    expected_uuid = '11111111-0000-4000-9009-002001001001'
    ac_2_stm = this_system_dir / 'ac-2_stmt.a'
    test_provided = ac_2_stm / f'{expected_uuid}.md'
    assert test_provided.exists()

    # confirm content in yaml header
    md_api = MarkdownAPI()
    header, tree = md_api.processor.process_markdown(test_provided)
    assert tree is not None

    comp_header_value = header[const.TRESTLE_LEVERAGING_COMP_TAG]
    assert comp_header_value == [{'name': 'REPLACE_ME'}]
    assert header[const.TRESTLE_STATEMENT_TAG][const.RESPONSIBILITY_UUID] == expected_uuid

    expected_responsibility = """# Responsibility Statement Description

Leveraging system's responsibilities with respect to inheriting this capability.

In the context of the application component in satisfaction of AC-2, part a.
"""

    # Confirm markdown content
    node = tree.get_node_for_key(const.RESPONSIBILITY_STATEMENT_DESCRIPTION, False)
    assert node.content.raw_text == expected_responsibility

    # Fine export files under Application
    application_dir = tmp_trestle_dir / ssp_name / const.INHERITANCE_VIEW_DIR / 'Application'

    expected_provided_uuid = '11111111-0000-4000-9009-002001002001'
    expected_responsibility_uuid = '11111111-0000-4000-9009-002001002002'
    ac_2_stm = application_dir / 'ac-2_stmt.a'
    test_provided = ac_2_stm / f'{expected_provided_uuid}_{expected_responsibility_uuid}.md'
    assert test_provided.exists()

    # confirm content in yaml header
    md_api = MarkdownAPI()
    header, tree = md_api.processor.process_markdown(test_provided)
    assert tree is not None

    comp_header_value = header[const.TRESTLE_LEVERAGING_COMP_TAG]
    assert comp_header_value == [{'name': 'REPLACE_ME'}]
    assert header[const.TRESTLE_STATEMENT_TAG][const.PROVIDED_UUID] == expected_provided_uuid
    assert header[const.TRESTLE_STATEMENT_TAG][const.RESPONSIBILITY_UUID] == expected_responsibility_uuid

    expected_provided = """# Provided Statement Description

Consumer-appropriate description of what may be inherited.

In the context of the application component in satisfaction of AC-2, part a.
"""

    expected_responsibility = """# Responsibility Statement Description

Leveraging system's responsibilities with respect to inheriting this capability.

In the context of the application component in satisfaction of AC-2, part a.
"""

    # Confirm markdown content
    node = tree.get_node_for_key(const.PROVIDED_STATEMENT_DESCRIPTION, False)
    assert node.content.raw_text == expected_provided

    node = tree.get_node_for_key(const.RESPONSIBILITY_STATEMENT_DESCRIPTION, False)
    assert node.content.raw_text == expected_responsibility


def test_ssp_assemble(tmp_trestle_dir: pathlib.Path) -> None:
    """Test ssp assemble from cli."""
    gen_args, _ = setup_for_ssp(tmp_trestle_dir, prof_name, ssp_name)
    args_compdefs = gen_args.compdefs

    # first create the markdown
    ssp_gen = SSPGenerate()
    assert ssp_gen._run(gen_args) == 0
    new_version = '1.2.3'

    prose_sys = 'My response for This System'
    prose_aa = 'My response for comp aa'
    prose_aa_a = 'My response for comp aa part a.'

    # edit it a bit
    ac_1_path = tmp_trestle_dir / ssp_name / 'ac/ac-1.md'
    assert test_utils.substitute_text_in_file(
        ac_1_path, '<!-- Add implementation prose for the main This System component for control: ac-1 -->', prose_sys
    )
    assert test_utils.substitute_text_in_file(ac_1_path, 'imp req prose for ac-1 from comp aa', prose_aa)
    assert test_utils.substitute_text_in_file(ac_1_path, 'statement prose for part a. from comp aa', prose_aa_a)
    # change status for sys comp
    assert test_utils.substitute_text_in_file(ac_1_path, 'Status: planned', 'Status: alternative')

    add_prompt = 'statement prose for part a. from comp ba'
    ac_67_path = tmp_trestle_dir / ssp_name / 'ac/ac-6.7.md'
    assert test_utils.substitute_text_in_file(ac_67_path, add_prompt, prose_aa_a)

    # generate markdown again on top of previous markdown to make sure it is not removed
    ssp_gen = SSPGenerate()
    assert ssp_gen._run(gen_args) == 0

    assert test_utils.replace_line_in_file_after_tag(
        ac_1_path, 'ac-1_prm_2:', '    values:\n    ssp-values:\n      - my ssp val\n'
    )
    assert test_utils.replace_line_in_file_after_tag(
        ac_1_path, '- shared_param_1_aa_opt_1', '      ssp-values:\n        - shared_param_1_aa_opt_1\n  comp_ab:\n'
    )

    # now assemble the edited controls into json ssp
    ssp_assemble = SSPAssemble()
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir,
        markdown=ssp_name,
        output=ssp_name,
        verbose=0,
        regenerate=False,
        version=new_version,
        name=None,
        compdefs=args_compdefs
    )
    assert ssp_assemble._run(args) == 0

    orig_ssp, orig_ssp_path = ModelUtils.load_model_for_class(tmp_trestle_dir, ssp_name, ossp.SystemSecurityPlan)
    orig_uuid = orig_ssp.uuid
    assert len(orig_ssp.system_implementation.components) == 5
    assert orig_ssp.metadata.version == new_version
    assert ModelUtils.model_age(orig_ssp) < test_utils.NEW_MODEL_AGE_SECONDS
    imp_reqs = orig_ssp.control_implementation.implemented_requirements
    imp_req = next((i_req for i_req in imp_reqs if i_req.control_id == 'ac-6.7'), None)
    assert imp_req.statements[0].by_components[0].description == prose_aa_a

    assert imp_reqs[0].by_components[0].set_parameters[1].param_id == 'shared_param_1'
    assert imp_reqs[0].by_components[0].set_parameters[1].values[0] == 'shared_param_1_aa_opt_1'
    assert imp_reqs[0].set_parameters[0].values[0] == 'my ssp val'

    by_comp = orig_ssp.control_implementation.implemented_requirements[0].by_components[2]
    assert by_comp.description == prose_sys

    orig_file_creation = orig_ssp_path.stat().st_mtime

    # now write it back out and confirm text is still there
    assert ssp_gen._run(gen_args) == 0
    assert confirm_control_contains(tmp_trestle_dir, 'ac-1', 'a.', prose_aa_a)
    assert test_utils.confirm_text_in_file(ac_1_path, const.SSP_MD_IMPLEMENTATION_QUESTION, prose_sys)

    # now assemble it again but don't regen uuid's and don't change version
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir,
        markdown=ssp_name,
        output=ssp_name,
        verbose=0,
        regenerate=False,
        name=None,
        version=None,
        compdefs=args_compdefs
    )
    assert ssp_assemble._run(args) == 0

    # confirm the file was not written out since no change
    assert orig_ssp_path.stat().st_mtime == orig_file_creation

    repeat_ssp, _ = ModelUtils.load_model_for_class(tmp_trestle_dir, ssp_name, ossp.SystemSecurityPlan)
    assert len(repeat_ssp.system_implementation.components) == 5
    assert repeat_ssp.metadata.version == new_version

    # assemble it again but regen uuid's
    # this should not regen uuid's because the file is not written out if only difference is uuid's
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir,
        markdown=ssp_name,
        output=ssp_name,
        verbose=0,
        regenerate=True,
        name=None,
        version=None,
        compdefs=args_compdefs
    )
    assert ssp_assemble._run(args) == 0
    assert orig_uuid == test_utils.get_model_uuid(tmp_trestle_dir, ssp_name, ossp.SystemSecurityPlan)
    # confirm the file was not written out since no change
    assert orig_ssp_path.stat().st_mtime == orig_file_creation
    # assemble it again but give new version and regen uuid's
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir,
        markdown=ssp_name,
        output=ssp_name,
        verbose=0,
        regenerate=True,
        name=None,
        version='new version to force write',
        compdefs=args_compdefs
    )
    assert ssp_assemble._run(args) == 0
    assert orig_uuid != test_utils.get_model_uuid(tmp_trestle_dir, ssp_name, ossp.SystemSecurityPlan)
    # confirm the file was not written out since no change
    assert orig_ssp_path.stat().st_mtime > orig_file_creation


def test_ssp_assemble_fedramp_profile(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Tests ssp assemble with a fedramp profile."""
    gen_args = setup_for_ssp_fedramp(tmp_trestle_dir, ssp_name)
    ssp_gen = SSPGenerate()
    assert ssp_gen._run(gen_args) == 0

    # first assemble
    ssp_assemble = f'trestle author ssp-assemble -m {ssp_name} -o {ssp_name} -cd {gen_args.compdefs}'
    test_utils.execute_command_and_assert(ssp_assemble, 0, monkeypatch)


def test_ssp_assemble_remove_comp_defs(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Tests the removal of component definitions that are no longer valid for an ssp."""
    gen_args, _ = setup_for_ssp(tmp_trestle_dir, prof_name, ssp_name)
    # first create the markdown
    ssp_gen = SSPGenerate()
    assert ssp_gen._run(gen_args) == 0

    # first assemble
    ssp_assemble = f'trestle author ssp-assemble -m {ssp_name} -o {ssp_name} -cd {gen_args.compdefs}'
    test_utils.execute_command_and_assert(ssp_assemble, 0, monkeypatch)
    # modify component uuids for testing removal
    orig_ssp, _ = ModelUtils.load_model_for_class(tmp_trestle_dir, ssp_name, ossp.SystemSecurityPlan)
    imp_reqs = orig_ssp.control_implementation.implemented_requirements
    components = orig_ssp.system_implementation.components
    generic_uuid = '46b7a556-72bb-4281-b805-a8f4030ca0e3'
    new_component = gens.generate_sample_model(ossp.SystemComponent)
    new_component.uuid = generic_uuid
    new_component.title = 'foo'
    components.append(new_component)
    by_comp = gens.generate_sample_model(ossp.ByComponent)
    by_comp.component_uuid = generic_uuid
    imp_reqs[0].by_components.append(by_comp)

    ModelUtils.save_top_level_model(orig_ssp, tmp_trestle_dir, ssp_name, FileContentType.JSON)
    # reassemble with changes
    ssp_assemble = f'trestle author ssp-assemble -m {ssp_name} -o {ssp_name} -cd {gen_args.compdefs}'
    test_utils.execute_command_and_assert(ssp_assemble, 0, monkeypatch)

    # loads edited ssp again
    edited_ssp, _ = ModelUtils.load_model_for_class(tmp_trestle_dir, ssp_name, ossp.SystemSecurityPlan)
    components = edited_ssp.system_implementation.components
    imp_reqs = edited_ssp.control_implementation.implemented_requirements
    assert not [x for x in components if x.uuid == generic_uuid]
    assert not [x for x in imp_reqs[0].by_components if x.component_uuid == generic_uuid]


def test_ssp_generate_bad_name(tmp_trestle_dir: pathlib.Path) -> None:
    """Test bad output name."""
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir, profile=prof_name, output='catalogs', verbose=0, yaml_header='dummy.yaml'
    )
    ssp_cmd = SSPGenerate()
    assert ssp_cmd._run(args) == 1


def test_ssp_generate_resolved_catalog(tmp_trestle_dir: pathlib.Path) -> None:
    """Test the ssp generator to create a resolved profile catalog."""
    _, _ = setup_for_ssp(tmp_trestle_dir, prof_name, ssp_name)
    profile_path = tmp_trestle_dir / f'profiles/{prof_name}/profile.json'
    new_catalog_dir = tmp_trestle_dir / f'catalogs/{prof_name}_resolved_catalog'
    new_catalog_dir.mkdir(parents=True, exist_ok=True)
    new_catalog_path = new_catalog_dir / 'catalog.json'

    profile_resolver = ProfileResolver()
    resolved_catalog = profile_resolver.get_resolved_profile_catalog(tmp_trestle_dir, profile_path)
    assert resolved_catalog
    # FIXME this should test with a more complex catalog
    assert len(resolved_catalog.groups) == 2

    resolved_catalog.oscal_write(new_catalog_path)


def test_ssp_assemble_with_inheritance(tmp_trestle_dir: pathlib.Path) -> None:
    """Test ssp assemble from cli with inheritance view."""
    gen_args, _ = setup_for_ssp(tmp_trestle_dir, prof_name, ssp_name, False, 'leveraged_ssp')
    args_compdefs = gen_args.compdefs

    # first create the markdown
    ssp_gen = SSPGenerate()
    assert ssp_gen._run(gen_args) == 0

    this_system_dir = tmp_trestle_dir / ssp_name / const.INHERITANCE_VIEW_DIR / 'This System'

    expected_uuid = '11111111-0000-4000-9009-001001002001'
    ac_21 = this_system_dir / 'ac-2.1'
    test_provided = ac_21 / f'{expected_uuid}.md'

    test_utils.replace_in_file(test_provided, 'REPLACE_ME', 'comp_aa')

    # now assemble the edited controls into json ssp
    ssp_assemble = SSPAssemble()
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir,
        markdown=ssp_name,
        output=ssp_name,
        verbose=0,
        regenerate=False,
        name=None,
        compdefs=args_compdefs,
        version=None
    )
    assert ssp_assemble._run(args) == 0

    ssp, _ = ModelUtils.load_model_for_class(tmp_trestle_dir, ssp_name, ossp.SystemSecurityPlan, FileContentType.JSON)

    imp_reqs = ssp.control_implementation.implemented_requirements
    imp_req = next((i_req for i_req in imp_reqs if i_req.control_id == 'ac-2.1'), None)
    inherited = imp_req.by_components[1].inherited[0]  # type: ignore
    assert inherited.description == (
        'Consumer-appropriate description of what may be inherited.\n\n\
In the context of the application component in satisfaction of AC-2.1.'
    )
    assert inherited.provided_uuid == expected_uuid


def test_ssp_filter(tmp_trestle_dir: pathlib.Path) -> None:
    """Test the ssp filter."""
    # FIXME enhance coverage
    # install the catalog and profiles
    gen_args, _ = setup_for_ssp(tmp_trestle_dir, prof_name, ssp_name)
    ssp_gen = SSPGenerate()
    assert ssp_gen._run(gen_args) == 0

    # create ssp from the markdown
    ssp_assemble = SSPAssemble()
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir,
        markdown=ssp_name,
        output=ssp_name,
        verbose=0,
        name=None,
        version=None,
        regenerate=False,
        compdefs=gen_args.compdefs
    )
    assert ssp_assemble._run(args) == 0

    ssp: ossp.SystemSecurityPlan
    ssp, _ = ModelUtils.load_model_for_class(tmp_trestle_dir, ssp_name, ossp.SystemSecurityPlan, FileContentType.JSON)

    assert len(ssp.control_implementation.implemented_requirements) == 8

    filtered_name = 'filtered_ssp'

    # now filter the ssp through comp_prof_aa
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir,
        name=ssp_name,
        profile='comp_prof_aa',
        output=filtered_name,
        verbose=0,
        regenerate=False,
        version=None,
        components=None,
        implementation_status=None,
        control_origination=None
    )
    ssp_filter = SSPFilter()
    assert ssp_filter._run(args) == 0

    ssp, _ = ModelUtils.load_model_for_class(
        tmp_trestle_dir,
        filtered_name,
        ossp.SystemSecurityPlan,
        FileContentType.JSON
    )

    # confirm the imp_reqs have been culled by profile_d to only two controls
    assert len(ssp.control_implementation.implemented_requirements) == 3

    # confirm there are three by_comps for: this system, foo, bar
    assert len(ssp.control_implementation.implemented_requirements[0].statements[0].by_components) == 2

    # now filter the ssp by components
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir,
        name=ssp_name,
        profile=None,
        output=filtered_name,
        verbose=0,
        regenerate=True,
        version=None,
        components='comp_aa',
        implementation_status=None,
        control_origination=None
    )
    ssp_filter = SSPFilter()
    assert ssp_filter._run(args) == 0

    ssp, _ = ModelUtils.load_model_for_class(
        tmp_trestle_dir,
        filtered_name,
        ossp.SystemSecurityPlan,
        FileContentType.JSON
    )

    # filter the filtered ssp again to confirm uuid does not change even with regen because contents are the same
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir,
        name=filtered_name,
        profile=None,
        output=filtered_name,
        verbose=0,
        regenerate=True,
        version=None,
        components='comp_aa',
        implementation_status=None,
        control_origination=None
    )
    ssp_filter = SSPFilter()
    assert ssp_filter._run(args) == 0

    # now filter the ssp by multiple implementation statuses
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir,
        name=ssp_name,
        profile=None,
        output=filtered_name,
        verbose=0,
        regenerate=False,
        version=None,
        components=None,
        implementation_status='not-applicable,implemented',
        control_origination=None
    )
    ssp_filter = SSPFilter()
    assert ssp_filter._run(args) == 0

    ssp, _ = ModelUtils.load_model_for_class(
        tmp_trestle_dir,
        filtered_name,
        ossp.SystemSecurityPlan,
        FileContentType.JSON
    )

    # confirm the imp_reqs have been culled by impl_status to five controls
    assert len(ssp.control_implementation.implemented_requirements) == 5
    # confirm there are is two by_comps for the first impl_req
    assert len(ssp.control_implementation.implemented_requirements[0].by_components) == 2

    # now filter the ssp by an implementation status that is unused
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir,
        name=ssp_name,
        profile=None,
        output=filtered_name,
        verbose=0,
        regenerate=False,
        version=None,
        components=None,
        implementation_status='not-applicable',
        control_origination=None
    )
    ssp_filter = SSPFilter()
    assert ssp_filter._run(args) == 0

    ssp, _ = ModelUtils.load_model_for_class(
        tmp_trestle_dir,
        filtered_name,
        ossp.SystemSecurityPlan,
        FileContentType.JSON
    )

    # confirm the imp_reqs have been culled by impl_status to zero controls
    assert len(ssp.control_implementation.implemented_requirements) == 0

    # now filter without profile or components to trigger error
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir,
        name=ssp_name,
        profile=None,
        output=filtered_name,
        verbose=0,
        regenerate=True,
        version=None,
        components=None,
        implementation_status=None,
        control_origination=None
    )

    ssp_filter = SSPFilter()
    assert ssp_filter._run(args) == 1

    # now filter the ssp through comp_prof_bad to force error because it references a control not in the ssp
    bad_prof = 'comp_prof_bad'
    test_utils.load_from_json(tmp_trestle_dir, bad_prof, bad_prof, prof.Profile)
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir,
        name=ssp_name,
        profile=bad_prof,
        output=filtered_name,
        verbose=0,
        regenerate=True,
        version=None,
        components=None,
        implementation_status=None,
        control_origination=None
    )
    ssp_filter = SSPFilter()
    assert ssp_filter._run(args) == 1

    # now filter with an invalid implementation status to trigger error
    bad_impl = 'impl_bad'
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir,
        name=ssp_name,
        profile=None,
        output=filtered_name,
        verbose=0,
        regenerate=True,
        version=None,
        components=None,
        implementation_status=bad_impl,
        control_origination=None
    )
    ssp_filter = SSPFilter()
    assert ssp_filter._run(args) == 1


def test_ssp_filter_control_origination(tmp_trestle_dir: pathlib.Path) -> None:
    """Test the ssp filter when filtering by control origination."""
    gen_args, _ = setup_for_ssp(tmp_trestle_dir, prof_name, ssp_name)
    ssp_gen = SSPGenerate()
    assert ssp_gen._run(gen_args) == 0

    # create ssp from the markdown
    ssp_assemble = SSPAssemble()
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir,
        markdown=ssp_name,
        output=ssp_name,
        verbose=0,
        name=None,
        version=None,
        regenerate=False,
        compdefs=gen_args.compdefs
    )
    assert ssp_assemble._run(args) == 0

    ssp: ossp.SystemSecurityPlan
    ssp, _ = ModelUtils.load_model_for_class(tmp_trestle_dir, ssp_name, ossp.SystemSecurityPlan, FileContentType.JSON)

    assert len(ssp.control_implementation.implemented_requirements) == 8

    filtered_name = 'filtered_ssp'

    # now filter the ssp by multiple control origination values
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir,
        name=ssp_name,
        profile=None,
        output=filtered_name,
        verbose=0,
        regenerate=False,
        version=None,
        components=None,
        implementation_status=None,
        control_origination='customer-configured,system-specific'
    )
    ssp_filter = SSPFilter()
    assert ssp_filter._run(args) == 0

    ssp, _ = ModelUtils.load_model_for_class(
        tmp_trestle_dir,
        filtered_name,
        ossp.SystemSecurityPlan,
        FileContentType.JSON
    )

    # confirm the imp_reqs have been culled to two controls
    assert len(ssp.control_implementation.implemented_requirements) == 2

    # now filter the ssp by a control origination that is unused
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir,
        name=ssp_name,
        profile=None,
        output=filtered_name,
        verbose=0,
        regenerate=False,
        version=None,
        components=None,
        implementation_status=None,
        control_origination='inherited'
    )
    ssp_filter = SSPFilter()
    assert ssp_filter._run(args) == 0

    ssp, _ = ModelUtils.load_model_for_class(
        tmp_trestle_dir,
        filtered_name,
        ossp.SystemSecurityPlan,
        FileContentType.JSON
    )

    # confirm the imp_reqs have been culled to zero controls
    assert len(ssp.control_implementation.implemented_requirements) == 0

    # filter with an invalid control origination to trigger error
    bad_co = 'co_bad'
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir,
        name=ssp_name,
        profile=None,
        output=filtered_name,
        verbose=0,
        regenerate=True,
        version=None,
        components=None,
        implementation_status=None,
        control_origination=bad_co
    )
    ssp_filter = SSPFilter()
    assert ssp_filter._run(args) == 1


def test_ssp_bad_control_id(tmp_trestle_dir: pathlib.Path) -> None:
    """Test ssp gen when profile has bad control id."""
    profile = prof.Profile.oscal_read(test_utils.JSON_TEST_DATA_PATH / 'profile_bad_control.json')
    ModelUtils.save_top_level_model(profile, tmp_trestle_dir, 'bad_prof', FileContentType.JSON)
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir, profile='bad_prof', output='my_ssp', verbose=0, sections=None, yaml_header=None
    )
    ssp_cmd = SSPGenerate()
    assert ssp_cmd._run(args) == 1


def test_ssp_assemble_header_metadata(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test parsing of metadata from yaml header."""
    args, _ = setup_for_ssp(tmp_trestle_dir, prof_name, ssp_name)
    header_path = test_utils.YAML_TEST_DATA_PATH / 'header_with_metadata.yaml'
    args.yaml_header = header_path
    ssp_cmd = SSPGenerate()
    assert ssp_cmd._run(args) == 0

    # create ssp from the markdown
    ssp_assemble = f'trestle author ssp-assemble -m {ssp_name} -o {ssp_name} -cd {args.compdefs}'
    test_utils.execute_command_and_assert(ssp_assemble, 0, monkeypatch)

    # read the assembled ssp and confirm roles are in metadata
    ssp, _ = ModelUtils.load_model_for_class(tmp_trestle_dir, ssp_name, ossp.SystemSecurityPlan, FileContentType.JSON)
    # FIXME assert len(ssp.metadata.roles) == 2


def test_ssp_force_overwrite(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test ssp generate with force-overwrite."""
    args, _ = setup_for_ssp(tmp_trestle_dir, prof_name, ssp_name)

    # confirm that setting force_overwrite with empty dir does not fail on generate
    args.force_overwrite = True
    ssp_cmd = SSPGenerate()
    assert ssp_cmd._run(args) == 0
    fc = test_utils.FileChecker(tmp_trestle_dir / 'my_ssp/')

    ac_1 = tmp_trestle_dir / ssp_name / 'ac' / 'ac-1.md'
    assert ac_1.exists()

    md_api = MarkdownAPI()
    header, tree = md_api.processor.process_markdown(ac_1)

    assert tree
    old_value = '<!-- Add implementation prose for the main This System component for control: ac-1 -->'
    assert old_value in tree.content.raw_text
    tree.content.raw_text = tree.content.raw_text.replace(old_value, 'Custom control implementation')
    md_api.write_markdown_with_header(ac_1, header, tree.content.raw_text)

    # re-run without force overwrite and confirm edits still there
    args.force_overwrite = False
    assert ssp_cmd._run(args) == 0

    header, tree = md_api.processor.process_markdown(ac_1)
    assert 'Custom control implementation' in tree.content.raw_text

    # run again with overwrite and confirm edits are gone
    args.force_overwrite = True
    assert ssp_cmd._run(args) == 0

    header, tree = md_api.processor.process_markdown(ac_1)
    assert old_value in tree.content.raw_text

    assert fc.files_unchanged()


def test_merge_statement() -> None:
    """Test merge statement."""
    imp_req = gens.generate_sample_model(ossp.ImplementedRequirement, True)
    new_statement = gens.generate_sample_model(ossp.Statement, True)
    prose = 'my new prose'
    new_statement.by_components[0].description = prose
    SSPAssemble._merge_statement(imp_req, new_statement, [])
    assert imp_req.statements[0].by_components[0].description == prose


def test_merge_imp_req() -> None:
    """Test merge statement."""
    imp_req_a = gens.generate_sample_model(ossp.ImplementedRequirement, True)
    imp_req_b = generic.GenericImplementedRequirement.generate()
    statement = generic.GenericStatement(statement_id=const.REPLACE_ME, uuid=const.SAMPLE_UUID_STR, description='foo')
    by_comp = generic.GenericByComponent.generate()
    prose = 'my new prose'
    by_comp.description = prose
    statement.by_components = [by_comp]
    imp_req_b.statements = [statement]
    SSPAssemble._merge_imp_req_into_imp_req(imp_req_a, imp_req_b, [])
    assert imp_req_a.statements[0].by_components[0].description == prose


def test_ssp_warning_missing_control(tmp_trestle_dir: pathlib.Path, capsys) -> None:
    """Test ssp success when profile missing control."""
    gen_args, _ = setup_for_ssp(tmp_trestle_dir, prof_name, ssp_name)
    prof_path = tmp_trestle_dir / 'profiles/comp_prof/profile.json'
    # remove the reference to control ac-1
    test_utils.delete_line_in_file(prof_path, 'ac-1')
    ssp_gen = SSPGenerate()
    assert ssp_gen._run(gen_args) == 0
    _, err = capsys.readouterr()
    assert 'Component comp_aa references control ac-1 not in profile.' in err


def test_ssp_assemble_no_comps(tmp_trestle_dir: pathlib.Path, capsys) -> None:
    """Test ssp assemble with no compdefs."""
    gen_args, _ = setup_for_ssp(tmp_trestle_dir, prof_name, ssp_name)
    gen_args.compdefs = None
    gen_args.yaml_header = None

    ssp_gen = SSPGenerate()
    assert ssp_gen._run(gen_args) == 0

    prose_sys = 'My response for This System'

    # edit it a bit
    ac_1_path = tmp_trestle_dir / ssp_name / 'ac/ac-1.md'
    assert test_utils.substitute_text_in_file(
        ac_1_path, '<!-- Add implementation prose for the main This System component for control: ac-1 -->', prose_sys
    )
    # change status for sys comp
    assert test_utils.substitute_text_in_file(ac_1_path, 'Status: planned', 'Status: alternative')

    # now assemble the edited controls into json ssp
    ssp_assemble = SSPAssemble()
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir,
        markdown=ssp_name,
        output=ssp_name,
        verbose=0,
        regenerate=False,
        version=None,
        name=None,
        compdefs=None
    )
    assert ssp_assemble._run(args) == 0

    assem_ssp, _ = ModelUtils.load_model_for_class(tmp_trestle_dir, ssp_name, ossp.SystemSecurityPlan)
    assert len(assem_ssp.system_implementation.components) == 1
    # following tests pass on windows but not others
    imp_req = list_utils.get_item_from_list(
        assem_ssp.control_implementation.implemented_requirements, 'ac-1', lambda x: x.control_id
    )
    by_comp = imp_req.by_components[0]
    assert by_comp.description == prose_sys
    assert by_comp.implementation_status.state == 'alternative'

    assert test_utils.replace_line_in_file_after_tag(
        ac_1_path, 'Status: alternative', '\n### Bad Component\n\n#### Status planned\n'
    )
    assert ssp_assemble._run(args) == 1
    _, err = capsys.readouterr()
    assert 'Control ac-1 references component Bad Component not defined' in err
    assert 'Please specify the names of any component-definitions' in err


def test_ssp_gen_and_assemble_more_than_one_param(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test ssp generate and assemble with more than 1 parameters per rule."""
    gen_args, _ = setup_for_ssp(tmp_trestle_dir, prof_name, ssp_name, False, '', 'comp_def_more_params')
    args_compdefs = gen_args.compdefs

    # first create the markdown
    ssp_gen = SSPGenerate()
    assert ssp_gen._run(gen_args) == 0
    new_version = '1.2.3'

    md_path = tmp_trestle_dir / ssp_name / 'ac' / 'ac-1.md'
    assert md_path.exists()

    md_api = MarkdownAPI()
    header, tree = md_api.processor.process_markdown(md_path)
    rule_parameters = header['x-trestle-comp-def-rules-param-vals']['comp_ca']
    rule_parameters.append({'name': 'allowed_admins_per_account2', 'values': ['20']})

    md_api.write_markdown_with_header(md_path, header, tree.content.raw_text)

    # verifies a second parameter has beend added to the top shared rule
    assert header['x-trestle-rules-params']['comp_ca'][1]['name'] == 'allowed_admins_per_account2'

    # verifies the parameter value for the rule has been written down correctly in the markdown file
    assert header['x-trestle-comp-def-rules-param-vals']['comp_ca'][1]['values'] == ['20']

    # now assemble controls into json ssp
    ssp_assemble = SSPAssemble()
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir,
        markdown=ssp_name,
        output=ssp_name,
        verbose=0,
        regenerate=False,
        version=new_version,
        name=None,
        compdefs=args_compdefs
    )
    assert ssp_assemble._run(args) == 0

    assem_ssp, _ = ModelUtils.load_model_for_class(tmp_trestle_dir, ssp_name, ossp.SystemSecurityPlan)
    set_parameters = assem_ssp.control_implementation.implemented_requirements[0].by_components[0].set_parameters
    set_params = [
        set_param.param_id for set_param in set_parameters if set_param.param_id == 'allowed_admins_per_account2'
    ]
    # this demonstrates there's only one iteration of the parameter and not being repeated
    assert len(set_params) == 1


def test_ssp_gen_throw_exception_for_rep_comps(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test ssp generate for duplicated component uuids between diff component definition."""
    gen_args, _ = setup_for_ssp(tmp_trestle_dir, prof_name, ssp_name, False, '',
                                'comp_def_more_params,comp_def_more_params_dup')
    # first create the markdown
    ssp_gen = SSPGenerate()
    assert ssp_gen._run(gen_args) == 1



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2023 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""cis-xlsx-to-oscal-catalog task tests."""

import configparser
import os
import pathlib
from typing import Dict
from unittest import mock

from openpyxl import load_workbook

import trestle.tasks.cis_xlsx_to_oscal_catalog as cis_xlsx_to_oscal_catalog
from trestle.oscal.catalog import Catalog
from trestle.tasks.base_task import TaskOutcome

ocp_config = 'tests/data/tasks/cis-xlsx-to-oscal-catalog/test-cis-xlsx-to-oscal-catalog.ocp.config'
rhel_config = 'tests/data/tasks/cis-xlsx-to-oscal-catalog/test-cis-xlsx-to-oscal-catalog.rhel.config'


def _get_section(tmp_path: pathlib.Path, file_: str) -> Dict:
    """Get section."""
    config = configparser.ConfigParser()
    config_path = pathlib.Path(file_)
    config.read(config_path)
    section = config['task.cis-xlsx-to-oscal-catalog']
    section['output-dir'] = str(tmp_path)
    return section


def test_cis_xlsx_to_oscal_catalog_print_info(tmp_path: pathlib.Path):
    """Test print_info call."""
    section = _get_section(tmp_path, ocp_config)
    tgt = cis_xlsx_to_oscal_catalog.CisXlsxToOscalCatalog(section)
    retval = tgt.print_info()
    assert retval is None


def test_cis_xlsx_to_oscal_catalog_simulate(tmp_path: pathlib.Path):
    """Test simulate call."""
    section = _get_section(tmp_path, ocp_config)
    tgt = cis_xlsx_to_oscal_catalog.CisXlsxToOscalCatalog(section)
    retval = tgt.simulate()
    assert retval == TaskOutcome.SIM_SUCCESS
    assert len(os.listdir(str(tmp_path))) == 0


def test_cis_xlsx_to_oscal_catalog_execute_ocp(tmp_path: pathlib.Path):
    """Test execute call - ocp."""
    section = _get_section(tmp_path, ocp_config)
    tgt = cis_xlsx_to_oscal_catalog.CisXlsxToOscalCatalog(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.SUCCESS
    _validate_ocp(tmp_path)


def _validate_ocp(tmp_path: pathlib.Path):
    """Validate produced OSCAL for OCP catalog."""
    # read catalog
    file_path = tmp_path / 'catalog.json'
    catalog = Catalog.oscal_read(file_path)
    # spot check
    assert len(catalog.groups) == 1
    # group 0
    g0 = catalog.groups[0]
    assert g0.id == 'CIS-1'
    assert g0.title == 'Control Plane Components'
    assert len(g0.groups) == 2
    assert len(g0.props) == 1
    p0 = g0.props[0]
    assert p0.name == 'status'
    assert p0.value == 'draft'
    g00 = g0.groups[0]
    assert g00.id == 'CIS-1.1'
    assert g00.title == 'Master Node Configuration Files'
    assert len(g0.props) == 1
    assert len(g00.controls) == 7
    c5 = g00.controls[5]
    p50 = c5.props[0]
    assert p50.name == 'profile'
    assert p50.value == 'Level 1'
    p52 = c5.props[2]
    assert p52.name == 'assessment_status'
    assert p52.value == 'Manual'


def test_cis_xlsx_to_oscal_catalog_no_overwrite(tmp_path: pathlib.Path):
    """Test execute call."""
    section = _get_section(tmp_path, ocp_config)
    tgt = cis_xlsx_to_oscal_catalog.CisXlsxToOscalCatalog(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.SUCCESS
    _validate_ocp(tmp_path)
    section['output-overwrite'] = 'False'
    tgt = cis_xlsx_to_oscal_catalog.CisXlsxToOscalCatalog(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_cis_xlsx_to_oscal_catalog_missing_config(tmp_path: pathlib.Path):
    """Test missing config."""
    section = None
    tgt = cis_xlsx_to_oscal_catalog.CisXlsxToOscalCatalog(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_cis_xlsx_to_oscal_catalog_missing_version(tmp_path: pathlib.Path):
    """Test missing version."""
    section = _get_section(tmp_path, ocp_config)
    section.pop('version')
    tgt = cis_xlsx_to_oscal_catalog.CisXlsxToOscalCatalog(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_cis_xlsx_to_oscal_catalog_missing_sheet(tmp_path: pathlib.Path):
    """Test missing sheet."""
    folder = 'tests/data/tasks/cis-xlsx-to-oscal-catalog'
    file_ = f'{folder}/CIS_RedHat_OpenShift_Container_Platform_Benchmark_v1.2.0-2.snippet.xlsx'
    wb_hacked = load_workbook(file_)
    sheet = wb_hacked['Combined Profiles']
    wb_hacked.remove(sheet)
    with mock.patch('trestle.tasks.cis_xlsx_to_oscal_catalog.load_workbook') as load_workbook_mock:
        load_workbook_mock.return_value = wb_hacked
        section = _get_section(tmp_path, ocp_config)
        tgt = cis_xlsx_to_oscal_catalog.CisXlsxToOscalCatalog(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.FAILURE


def test_cis_xlsx_to_oscal_catalog_one_dot_added_part(tmp_path: pathlib.Path):
    """Test group part."""
    folder = 'tests/data/tasks/cis-xlsx-to-oscal-catalog'
    file_ = f'{folder}/CIS_RedHat_OpenShift_Container_Platform_Benchmark_v1.2.0-2.snippet.xlsx'
    wb_hacked = load_workbook(file_)
    sheet = wb_hacked['Combined Profiles']
    cell = sheet.cell(2, 7)
    assert cell.value.startswith('This section consists of security recommendations for')
    cell = sheet.cell(3, 7)
    cell.value = 'foobar'
    with mock.patch('trestle.tasks.cis_xlsx_to_oscal_catalog.load_workbook') as load_workbook_mock:
        load_workbook_mock.return_value = wb_hacked
        section = _get_section(tmp_path, ocp_config)
        tgt = cis_xlsx_to_oscal_catalog.CisXlsxToOscalCatalog(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.SUCCESS
    # read catalog
    file_path = tmp_path / 'catalog.json'
    catalog = Catalog.oscal_read(file_path)
    # spot check
    g0 = catalog.groups[0]
    g00 = g0.groups[0]
    assert len(g00.parts) == 1
    p000 = g00.parts[0]
    assert p000.prose == 'foobar'


def test_cis_xlsx_to_oscal_catalog_unexpected_section(tmp_path: pathlib.Path):
    """Test group part."""
    folder = 'tests/data/tasks/cis-xlsx-to-oscal-catalog'
    file_ = f'{folder}/CIS_RedHat_OpenShift_Container_Platform_Benchmark_v1.2.0-2.snippet.xlsx'
    wb_hacked = load_workbook(file_)
    sheet = wb_hacked['Combined Profiles']
    cell = sheet.cell(3, 1)
    assert cell.value == '1.1'
    cell.value = '1.2.3.4.5.6.7.8.9.0'
    with mock.patch('trestle.tasks.cis_xlsx_to_oscal_catalog.load_workbook') as load_workbook_mock:
        load_workbook_mock.return_value = wb_hacked
        section = _get_section(tmp_path, ocp_config)
        tgt = cis_xlsx_to_oscal_catalog.CisXlsxToOscalCatalog(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.FAILURE


def test_cis_xlsx_to_oscal_catalog_execute_rhel(tmp_path: pathlib.Path):
    """Test execute call - rhel."""
    section = _get_section(tmp_path, rhel_config)
    tgt = cis_xlsx_to_oscal_catalog.CisXlsxToOscalCatalog(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.SUCCESS
    _validate_rhel(tmp_path)


def _validate_rhel(tmp_path: pathlib.Path):
    """Validate produced OSCAL for RHEL catalog."""
    # read catalog
    file_path = tmp_path / 'catalog.json'
    catalog = Catalog.oscal_read(file_path)
    # spot check
    assert len(catalog.groups) == 1
    # group 0
    g0 = catalog.groups[0]
    assert g0.id == 'CIS-1'
    assert g0.title == 'Initial Setup'
    assert len(g0.groups) == 1
    assert len(g0.parts) == 1
    assert len(g0.controls) == 2
    p00 = g0.parts[0]
    assert p00.id == 'CIS-1_smt'
    assert p00.name == 'statement'
    assert p00.prose.startswith('Items in this section are advised for all systems,')
    g00 = g0.groups[0]
    assert g00.id == 'CIS-1.1'
    assert g00.title == 'Filesystem Configuration'
    c00 = g0.controls[0]
    assert c00.id == 'CIS-1.9'
    assert c00.title == 'Ensure updates, patches, and additional security software are installed'
    assert len(c00.props) == 9
    prop0 = c00.props[0]
    assert prop0.name == 'profile'
    assert prop0.value == 'Level 1 - Server'
    assert len(c00.parts) == 9
    part0 = c00.parts[0]
    assert part0.id == 'CIS-1.9_smt'
    assert part0.name == 'statement'
    t1 = 'Periodically patches are released for included software either due'
    t2 = ' to security flaws or to include additional functionality.'
    assert part0.prose == t1 + t2
    md = catalog.metadata
    assert md.title == 'CIS Red Hat Enterprise Linux 8 Benchmark'
    rs = catalog.back_matter.resources
    assert len(rs) == 1
    assert rs[
        0].description == 'CRYPTO-POLICIES(7):https://access.redhat.com/articles/3642912#what-polices-are-provided-1'



# Copyright (c) 2021 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Provided interface to write provided and responsibility exports statements to Markdown."""

import logging
import pathlib
from typing import Dict

import trestle.common.const as const
import trestle.oscal.ssp as ossp
from trestle.common.err import TrestleError
from trestle.common.list_utils import as_list
from trestle.core.crm.bycomp_interface import ByComponentInterface
from trestle.core.crm.leveraged_statements import (
    LeveragedStatements,
    StatementProvided,
    StatementResponsibility,
    StatementTree,
)

logger = logging.getLogger(__name__)


class ExportWriter:
    """
    By-Component Assembly Exports writer.

    Export writer handles all operations related to writing provided and responsibility exported statements
    to Markdown.
    """

    def __init__(self, root_path: pathlib.Path, ssp: ossp.SystemSecurityPlan, leveraged_ssp_href: str):
        """
        Initialize export writer.

        Arguments:
            root_path: A root path object where all markdown files and directories should be written.
            ssp: A system security plan with exports
        """
        self._ssp: ossp.SystemSecurityPlan = ssp
        self._root_path: pathlib.Path = root_path
        self._leveraged_ssp_href: str = leveraged_ssp_href

        # Find all the components and create paths for name
        self._paths_by_comp: Dict[str, pathlib.Path] = {}
        for component in as_list(self._ssp.system_implementation.components):
            self._paths_by_comp[component.uuid] = self._root_path.joinpath(component.title)

    def write_exports_as_markdown(self) -> None:
        """Write export statement for leveraged SSP as the inheritance Markdown view."""
        # Process all information under exports in control implementation
        for implemented_requirement in as_list(self._ssp.control_implementation.implemented_requirements):
            for by_comp in as_list(implemented_requirement.by_components):
                self._process_by_component(by_comp, implemented_requirement.control_id)

            for stm in as_list(implemented_requirement.statements):
                statement_id = getattr(stm, 'statement_id', f'{implemented_requirement.control_id}_smt')
                for by_comp in as_list(stm.by_components):
                    self._process_by_component(by_comp, statement_id)

    def _process_by_component(self, by_comp: ossp.ByComponent, control_id: str) -> None:
        """Complete the Markdown writing operations for each by-component assembly."""
        if by_comp.component_uuid not in self._paths_by_comp:
            raise TrestleError(f'Component id {by_comp.component_uuid} is not in the system implementation')

        comp_path: pathlib.Path = self._paths_by_comp[by_comp.component_uuid]

        export_interface: ByComponentInterface = ByComponentInterface(by_comp=by_comp)
        leveraged_statements: Dict[str, LeveragedStatements] = self._statement_types_from_exports(export_interface)

        # Only create the directory if leveraged statements exist. If not return.
        if not leveraged_statements:
            logger.debug(f'Component {by_comp.component_uuid} has no exports for control {control_id}')
            return

        control_path: pathlib.Path = comp_path.joinpath(control_id)
        control_path.mkdir(exist_ok=True, parents=True)

        for statement_file_path, leveraged_stm in leveraged_statements.items():
            statement_path: pathlib.Path = control_path.joinpath(f'{statement_file_path}{const.MARKDOWN_FILE_EXT}')
            leveraged_stm.write_statement_md(statement_path)

    def _statement_types_from_exports(self, export_interface: ByComponentInterface) -> Dict[str, LeveragedStatements]:
        """Process all exports and return a file basename and LeveragedStatement object for each."""
        all_statements: Dict[str, LeveragedStatements] = {}

        for responsibility in export_interface.get_isolated_responsibilities():
            all_statements[responsibility.uuid] = StatementResponsibility(
                responsibility.uuid, responsibility.description, self._leveraged_ssp_href
            )

        for provided in export_interface.get_isolated_provided():
            all_statements[provided.uuid
                           ] = StatementProvided(provided.uuid, provided.description, self._leveraged_ssp_href)

        for responsibility, provided in export_interface.get_export_sets():
            path = f'{provided.uuid}_{responsibility.uuid}'
            all_statements[path] = StatementTree(
                provided.uuid,
                provided.description,
                responsibility.uuid,
                responsibility.description,
                self._leveraged_ssp_href
            )

        return all_statements



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the ExportWriter class."""

import os
import pathlib
from typing import List, Tuple
from unittest.mock import Mock, patch

import pytest

from tests import test_utils

import trestle.core.generators as gens
import trestle.oscal.ssp as ossp
from trestle.common.err import TrestleError
from trestle.common.model_utils import ModelUtils
from trestle.core.crm.bycomp_interface import ByComponentInterface
from trestle.core.crm.export_writer import ExportWriter
from trestle.core.crm.leveraged_statements import (
    LeveragedStatements,
    StatementProvided,
    StatementResponsibility,
    StatementTree,
)
from trestle.core.models.file_content_type import FileContentType

test_profile = 'simple_test_profile'
test_ssp = 'leveraged_ssp'
test_ref = 'trestle://leveraged_ssp.json'


def custom_side_effect(file_path: pathlib.Path) -> None:
    """Write a test file."""
    with open(file_path, 'w') as file:
        file.write('test')


def test_write_exports_as_markdown(tmp_trestle_dir: pathlib.Path) -> None:
    """Test happy path for writing markdown with a mock LeveragedStatement."""
    _ = test_utils.setup_for_inherit(tmp_trestle_dir, test_profile, '', test_ssp)
    ssp, _ = ModelUtils.load_model_for_class(tmp_trestle_dir, test_ssp, ossp.SystemSecurityPlan, FileContentType.JSON)

    inherited_path = tmp_trestle_dir.joinpath('inherited')
    writer = ExportWriter(inherited_path, ssp, test_ref)

    mock = Mock(spec=LeveragedStatements)
    mock.write_statement_md.side_effect = custom_side_effect

    with patch('trestle.core.crm.export_writer.ExportWriter._statement_types_from_exports') as mock_process:
        return_value: List[Tuple[str, LeveragedStatements]] = {'filepath': mock}
        mock_process.return_value = return_value
        writer.write_exports_as_markdown()

        assert os.path.exists(inherited_path.joinpath('This System', 'ac-2.1', 'filepath.md'))
        # Check that directory are not created when no exports exists
        assert not os.path.exists(inherited_path.joinpath('Application', 'ac2'))


def test_write_exports_as_markdown_invalid_ssp(tmp_trestle_dir: pathlib.Path) -> None:
    """Test triggering an error with an invalid SSP input."""
    _ = test_utils.setup_for_inherit(tmp_trestle_dir, test_profile, '', test_ssp)
    ssp, _ = ModelUtils.load_model_for_class(tmp_trestle_dir, test_ssp, ossp.SystemSecurityPlan, FileContentType.JSON)

    # Delete a component that is used to create an invalid SSP
    ssp.system_implementation.components.remove(ssp.system_implementation.components[2])

    inherited_path = tmp_trestle_dir.joinpath('inherited')
    writer = ExportWriter(inherited_path, ssp, test_ref)

    with pytest.raises(TrestleError, match=r'Component .* is not in the system implementation'):
        writer.write_exports_as_markdown()


def test_statement_types_from_exports(tmp_trestle_dir: pathlib.Path) -> None:
    """Test generated LeveragedStatements and filenames with SSP input."""
    expected_provided = 1
    expected_responsibility = 1
    expected_set = 1

    ssp = gens.generate_sample_model(ossp.SystemSecurityPlan)

    inherited_path = tmp_trestle_dir.joinpath('inherited')
    writer = ExportWriter(inherited_path, ssp, test_ref)

    by_comp: ossp.ByComponent = test_utils.generate_test_by_comp()
    bycomp_interface: ByComponentInterface = ByComponentInterface(by_comp)

    result_leveraged_statements = writer._statement_types_from_exports(bycomp_interface)
    provided: List[StatementProvided] = []
    responsibility: List[StatementResponsibility] = []
    sets: List[StatementTree] = []

    for leveraged_stm in result_leveraged_statements.values():
        if isinstance(leveraged_stm, StatementProvided):
            provided.append(leveraged_stm)
        elif isinstance(leveraged_stm, StatementResponsibility):
            responsibility.append(leveraged_stm)
        elif isinstance(leveraged_stm, StatementTree):
            sets.append(leveraged_stm)

    assert len(provided) == expected_provided
    assert len(responsibility) == expected_responsibility
    assert len(sets) == expected_set


def test_statement_types_no_exports(tmp_trestle_dir: pathlib.Path) -> None:
    """Test generated LeveragedStatements and filenames with no exports."""
    ssp = gens.generate_sample_model(ossp.SystemSecurityPlan)

    inherited_path = tmp_trestle_dir.joinpath('inherited')
    writer = ExportWriter(inherited_path, ssp, test_ref)

    by_comp = gens.generate_sample_model(ossp.ByComponent)
    bycomp_interface: ByComponentInterface = ByComponentInterface(by_comp)

    result_leveraged_statements = writer._statement_types_from_exports(bycomp_interface)

    assert len(result_leveraged_statements) == 0



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2021 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for control input output methods."""

import pathlib
from typing import Any, Dict

import trestle.common.const as const
from trestle.core.crm.leveraged_statements import (
    InheritanceMarkdownReader, StatementProvided, StatementResponsibility, StatementTree
)
from trestle.core.markdown.markdown_api import MarkdownAPI
from trestle.core.markdown.md_writer import MDWriter

provided_uuid = '18ac4e2a-b5f2-46e4-94fa-cc84ab6fe114'
provided_statement_desc = 'provided statement description'
resp_uuid = '4b34c68f-75fa-4b38-baf0-e50158c13ac2'
resp_statement_desc = 'resp statement description'
satisfied_statement_desc = 'satisfied statement description'

test_href = 'trestle://ssp/ssp.json'


def add_authored_content(test_file: pathlib.Path, yaml_header: Dict[str, Any]) -> None:
    """Update the yaml header with a test component and satisfied description to simulate editing."""
    md_writer = MDWriter(test_file)
    yaml_header[const.TRESTLE_LEVERAGING_COMP_TAG] = [{'name': 'My_Comp'}]
    md_writer.add_yaml_header(yaml_header)
    md_writer.new_header(level=1, title=const.SATISFIED_STATEMENT_DESCRIPTION)
    md_writer.new_line(const.SATISFIED_STATEMENT_COMMENT)
    md_writer.new_line('My Satisfied Description')
    md_writer.write_out()


def test_write_inheritance_tree(tmp_path: pathlib.Path) -> None:
    """Test writing statements with both provided and responsibility."""
    statement_tree_path = tmp_path.joinpath('statement_tree.md')

    statement = StatementTree(provided_uuid, provided_statement_desc, resp_uuid, resp_statement_desc, test_href)

    statement.write_statement_md(statement_tree_path)

    # confirm content in yaml header
    md_api = MarkdownAPI()
    header, tree = md_api.processor.process_markdown(statement_tree_path)
    assert tree is not None

    comp_header_value = header[const.TRESTLE_LEVERAGING_COMP_TAG]
    assert comp_header_value == [{'name': 'REPLACE_ME'}]

    assert header[const.TRESTLE_STATEMENT_TAG][const.PROVIDED_UUID] == provided_uuid
    assert header[const.TRESTLE_STATEMENT_TAG][const.RESPONSIBILITY_UUID] == resp_uuid
    assert header[const.TRESTLE_GLOBAL_TAG][const.LEVERAGED_SSP][const.HREF] == test_href

    # Confirm markdown content
    node = tree.get_node_for_key(const.PROVIDED_STATEMENT_DESCRIPTION, False)
    assert node.content.raw_text == '# Provided Statement Description\n\nprovided statement description\n'
    node = tree.get_node_for_key(const.RESPONSIBILITY_STATEMENT_DESCRIPTION, False)
    assert node.content.raw_text == '# Responsibility Statement Description\n\nresp statement description\n'
    node = tree.get_node_for_key(const.SATISFIED_STATEMENT_DESCRIPTION, False)
    assert node.content.raw_text == (
        '# Satisfied Statement Description\n\n<!-- Use this section to explain '
        'how the inherited responsibility is being satisfied. -->'
    )

    # Update the component mapping and run again to make sure it persists
    add_authored_content(statement_tree_path, header)

    statement.write_statement_md(statement_tree_path)

    # Reread the Markdown
    md_api = MarkdownAPI()
    header, tree = md_api.processor.process_markdown(statement_tree_path)

    # Ensure My_Comp and satisfied description persists
    comp_header_value = header[const.TRESTLE_LEVERAGING_COMP_TAG]
    assert comp_header_value == [{'name': 'My_Comp'}]
    node = tree.get_node_for_key(const.SATISFIED_STATEMENT_DESCRIPTION, False)
    assert node.content.raw_text == """# Satisfied Statement Description\n
<!-- Use this section to explain how the inherited responsibility is being satisfied. -->\nMy Satisfied Description"""


def test_write_inheritance_provided(tmp_path: pathlib.Path) -> None:
    """Test writing statements with only provided."""
    statement_provided_path = tmp_path.joinpath('statement_provided.md')

    statement = StatementProvided(provided_uuid, provided_statement_desc, test_href)

    statement.write_statement_md(statement_provided_path)

    # confirm content in yaml header
    md_api = MarkdownAPI()
    header, tree = md_api.processor.process_markdown(statement_provided_path)
    assert tree is not None

    comp_header_value = header[const.TRESTLE_LEVERAGING_COMP_TAG]
    assert comp_header_value == [{'name': 'REPLACE_ME'}]

    assert header[const.TRESTLE_STATEMENT_TAG][const.PROVIDED_UUID] == provided_uuid
    assert header[const.TRESTLE_GLOBAL_TAG][const.LEVERAGED_SSP][const.HREF] == test_href

    # Confirm markdown content
    node = tree.get_node_for_key(const.PROVIDED_STATEMENT_DESCRIPTION, False)
    assert node.content.raw_text == '# Provided Statement Description\n\nprovided statement description'

    # Update the component mapping and run again to make sure it persists
    add_authored_content(statement_provided_path, header)

    statement.write_statement_md(statement_provided_path)

    # Reread the markdown
    md_api = MarkdownAPI()
    header, tree = md_api.processor.process_markdown(statement_provided_path)

    # Ensure My_Comp persists
    comp_header_value = header[const.TRESTLE_LEVERAGING_COMP_TAG]
    assert comp_header_value == [{'name': 'My_Comp'}]
    node = tree.get_node_for_key(const.SATISFIED_STATEMENT_DESCRIPTION, False)
    assert node is None


def test_write_inheritance_responsibility(tmp_path: pathlib.Path) -> None:
    """Test writing statements with only responsibility."""
    statement_resp_path = tmp_path.joinpath('statement_req.md')

    statement = StatementResponsibility(resp_uuid, resp_statement_desc, test_href)

    statement.write_statement_md(statement_resp_path)

    # confirm content in yaml header
    md_api = MarkdownAPI()
    header, tree = md_api.processor.process_markdown(statement_resp_path)
    assert tree is not None

    comp_header_value = header[const.TRESTLE_LEVERAGING_COMP_TAG]
    assert comp_header_value == [{'name': 'REPLACE_ME'}]

    assert header[const.TRESTLE_STATEMENT_TAG][const.RESPONSIBILITY_UUID] == resp_uuid
    assert header[const.TRESTLE_GLOBAL_TAG][const.LEVERAGED_SSP][const.HREF] == test_href

    # Confirm markdown content
    node = tree.get_node_for_key(const.RESPONSIBILITY_STATEMENT_DESCRIPTION, False)
    assert node.content.raw_text == '# Responsibility Statement Description\n\nresp statement description\n'
    node = tree.get_node_for_key(const.SATISFIED_STATEMENT_DESCRIPTION, False)
    assert node.content.raw_text == """# Satisfied Statement Description\n
<!-- Use this section to explain how the inherited responsibility is being satisfied. -->"""

    # Update the component mapping and run again to make sure it persists
    add_authored_content(statement_resp_path, header)

    statement.write_statement_md(statement_resp_path)

    # Reread the Markdown
    md_api = MarkdownAPI()
    header, tree = md_api.processor.process_markdown(statement_resp_path)

    # Ensure My_Comp and satisfied description persists
    comp_header_value = header[const.TRESTLE_LEVERAGING_COMP_TAG]
    assert comp_header_value == [{'name': 'My_Comp'}]
    node = tree.get_node_for_key(const.SATISFIED_STATEMENT_DESCRIPTION, False)
    assert node.content.raw_text == """# Satisfied Statement Description\n
<!-- Use this section to explain how the inherited responsibility is being satisfied. -->\nMy Satisfied Description"""


def test_process_leveraged_statement_default_mapping(tmp_path: pathlib.Path) -> None:
    """Test processing leveraged statement markdown with no set mapping."""
    statement_tree_path = tmp_path.joinpath('statement_tree.md')

    statement = StatementTree(provided_uuid, provided_statement_desc, resp_uuid, resp_statement_desc, test_href)

    statement.write_statement_md(statement_tree_path)

    md_reader: InheritanceMarkdownReader = InheritanceMarkdownReader(statement_tree_path)

    leveraging_information = md_reader.process_leveraged_statement_markdown()

    assert leveraging_information is None


def test_process_leveraged_statement_markdown_tree(tmp_path: pathlib.Path) -> None:
    """Test processing a statement tree in Markdown."""
    statement_tree_path = tmp_path.joinpath('statement_tree.md')

    # Add test mapped component
    test_header: Dict[str, Any] = {}
    add_authored_content(statement_tree_path, test_header)

    statement = StatementTree(provided_uuid, provided_statement_desc, resp_uuid, resp_statement_desc, test_href)

    statement.write_statement_md(statement_tree_path)

    md_reader: InheritanceMarkdownReader = InheritanceMarkdownReader(statement_tree_path)

    leveraging_information = md_reader.process_leveraged_statement_markdown()

    assert leveraging_information

    assert 'My_Comp' in leveraging_information.leveraging_comp_titles

    assert leveraging_information.inherited is not None
    inherited = leveraging_information.inherited
    assert inherited.provided_uuid == provided_uuid
    assert inherited.description == provided_statement_desc

    assert leveraging_information.satisfied is not None
    satisfied = leveraging_information.satisfied
    assert satisfied.responsibility_uuid == resp_uuid
    assert satisfied.description == 'My Satisfied Description'


def test_process_leveraged_statement_markdown_provided(tmp_path: pathlib.Path) -> None:
    """Test processing a statement provided markdown."""
    statement_provided_path = tmp_path.joinpath('statement_provided.md')

    # Add test mapped component
    test_header: Dict[str, Any] = {}
    add_authored_content(statement_provided_path, test_header)

    statement = StatementProvided(provided_uuid, provided_statement_desc, test_href)

    statement.write_statement_md(statement_provided_path)

    md_reader: InheritanceMarkdownReader = InheritanceMarkdownReader(statement_provided_path)

    leveraging_information = md_reader.process_leveraged_statement_markdown()

    assert leveraging_information

    assert 'My_Comp' in leveraging_information.leveraging_comp_titles

    assert leveraging_information.inherited is not None
    inherited = leveraging_information.inherited
    assert inherited.provided_uuid == provided_uuid
    assert inherited.description == provided_statement_desc

    assert leveraging_information.satisfied is None


def test_process_leveraged_statement_markdown_responsibility(tmp_path: pathlib.Path) -> None:
    """Test processing a statement responsibility Markdown."""
    statement_resp_path = tmp_path.joinpath('statement_req.md')

    # Add test mapped component
    test_header: Dict[str, Any] = {}
    add_authored_content(statement_resp_path, test_header)

    statement = StatementResponsibility(resp_uuid, resp_statement_desc, test_href)

    statement.write_statement_md(statement_resp_path)

    md_reader: InheritanceMarkdownReader = InheritanceMarkdownReader(statement_resp_path)

    leveraging_information = md_reader.process_leveraged_statement_markdown()

    assert leveraging_information

    assert 'My_Comp' in leveraging_information.leveraging_comp_titles

    assert leveraging_information.inherited is None

    assert leveraging_information.satisfied is not None
    satisfied = leveraging_information.satisfied
    assert satisfied.responsibility_uuid == resp_uuid
    assert satisfied.description == 'My Satisfied Description'



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for models module."""

import pathlib

from ruamel.yaml import YAML

from trestle.common import const
from trestle.core import parser

yaml_path = pathlib.Path('tests/data/yaml/')


def test_parse_dict() -> None:
    """Test parse_dict."""
    file_name = 'good_component.yaml'

    with open(pathlib.Path.joinpath(yaml_path, file_name), 'r', encoding=const.FILE_ENCODING) as f:
        yaml = YAML(typ='safe')
        data = yaml.load(f)
        target = parser.parse_dict(data['component-definition'], 'trestle.oscal.component.ComponentDefinition')
        assert target is not None



# Copyright (c) 2021 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Create resolved catalog from profile."""

import logging
import pathlib
from typing import Any, Dict, List, Optional, Tuple

import trestle.oscal.catalog as cat
import trestle.oscal.common as com
import trestle.oscal.profile as prof
from trestle.common.const import TRESTLE_INHERITED_PROPS_TRACKER
from trestle.common.list_utils import as_list, pop_item_from_list
from trestle.core.catalog.catalog_interface import CatalogInterface
from trestle.core.control_interface import ParameterRep
from trestle.core.resolver._import import Import

logger = logging.getLogger(__name__)


class ProfileResolver():
    """Class to resolve a catalog given a profile."""

    @staticmethod
    def _extract_inherited_props(res_cat: cat.Catalog) -> Tuple[cat.Catalog, Dict[str, Any]]:
        """
        Build the control dict of inherited props.

        Args:
            The resolved profile catalog with a possible temporary part in each control

        Returns:
            The temporary parts are removed from each control and a Dict of added props per control_id is returned

        Notes:
        If an upstream profile adds props to the control they are tracked in a special temporary part in the control
        called const.TRESTLE_INHERITED_PROPS_TRACKER.  If that part is present in a control its contents should be added
        to the dict entry for that control, and the part removed from the control.
        """
        prop_dict: Dict[str, Any] = {}
        cat_interface = CatalogInterface(res_cat)
        for control in cat_interface.get_all_controls_from_dict():
            part: com.Part = pop_item_from_list(control.parts, TRESTLE_INHERITED_PROPS_TRACKER, lambda p: p.name)
            if part:
                props_list: List[Dict[str, Any]] = []
                for prop in as_list(part.props):
                    props_list.append({'name': prop.name, 'value': prop.value})
                for sub_part in as_list(part.parts):
                    for prop in as_list(sub_part.props):
                        props_list.append({'name': prop.name, 'value': prop.value, 'part_name': sub_part.title})
                prop_dict[control.id] = props_list
        cat_interface.update_catalog_controls()
        clean_res_cat = cat_interface.get_catalog()
        return clean_res_cat, prop_dict

    @staticmethod
    def get_resolved_profile_catalog_and_inherited_props(
        trestle_root: pathlib.Path,
        profile_path: str,
        block_adds: bool = False,
        block_params: bool = False,
        params_format: Optional[str] = None,
        param_rep: ParameterRep = ParameterRep.LEAVE_MOUSTACHE,
        show_value_warnings: bool = False,
        value_assigned_prefix: Optional[str] = None,
        value_not_assigned_prefix: Optional[str] = None
    ) -> Tuple[cat.Catalog, Optional[Dict[str, Any]]]:
        """
        Create the resolved profile catalog given a profile path along with inherited props.

        Args:
            trestle_root: root directory of the trestle workspace
            profile_path: string path or uri of the profile being resolved
            block_adds: prevent the application of adds in the final profile
            block_params: prevent the application of setparams in the final profile
            params_format: optional pattern with dot to wrap the param string, where dot represents the param string
            param_rep: desired way to convert params to strings
            show_value_warnings: warn if prose references a value that has not been set
            value_assigned_prefix: Prefix placed in front of param string if a value was assigned
            value_not_assigned_prefix: Prefix placed in front of param string if a value was *not* assigned

        Returns:
            The resolved profile catalog and a control dict of inherited props
        """
        logger.debug(f'get resolved profile catalog and inherited props for {profile_path} via generated Import.')
        import_ = prof.Import(href=str(profile_path), include_all={})
        # The final Import has change_prose=True to force parameter substitution in the prose only at the last stage.
        import_filter = Import(
            trestle_root,
            import_, [],
            True,
            block_adds,
            block_params,
            params_format,
            param_rep,
            None,
            show_value_warnings,
            value_assigned_prefix,
            value_not_assigned_prefix
        )
        logger.debug('launch pipeline')
        resolved_profile_catalog = next(import_filter.process())
        resolved_profile_catalog, inherited_props = ProfileResolver._extract_inherited_props(resolved_profile_catalog)
        return resolved_profile_catalog, inherited_props

    @staticmethod
    def get_resolved_profile_catalog(
        trestle_root: pathlib.Path,
        profile_path: str,
        block_adds: bool = False,
        block_params: bool = False,
        params_format: Optional[str] = None,
        param_rep: ParameterRep = ParameterRep.LEAVE_MOUSTACHE,
        show_value_warnings: bool = False,
        value_assigned_prefix: Optional[str] = None,
        value_not_assigned_prefix: Optional[str] = None
    ) -> cat.Catalog:
        """
        Create the resolved profile catalog given a profile path.

        Args:
            trestle_root: root directory of the trestle workspace
            profile_path: string path or uri of the profile being resolved
            block_adds: prevent the application of adds in the final profile
            block_params: prevent the application of setparams in the final profile
            params_format: optional pattern with dot to wrap the param string, where dot represents the param string
            param_rep: desired way to convert params to strings
            show_value_warnings: warn if prose references a value that has not been set
            value_assigned_prefix: Prefix placed in front of param string if a value was assigned
            value_not_assigned_prefix: Prefix placed in front of param string if a value was *not* assigned

        Returns:
            The resolved profile catalog
        """
        logger.debug(f'get resolved profile catalog for {profile_path} via generated Import.')
        resolved_profile_catalog, _ = ProfileResolver.get_resolved_profile_catalog_and_inherited_props(
            trestle_root,
            profile_path,
            block_adds,
            block_params,
            params_format,
            param_rep,
            show_value_warnings,
            value_assigned_prefix,
            value_not_assigned_prefix
        )
        return resolved_profile_catalog
